from typing import Dict

import click
from cognite.client.exceptions import CogniteAPIError

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.utils import exit_with_cognite_api_error, print_query


@click.command(help="Make a SQL query and retrieve results")
@click.argument("query")
@click.option(
    "--source-limit", default=100, help="This limits the number of rows to read from each data source, defaults to 100."
)
@click.option("--infer-schema-limit", default=100, help="Schema inference limit, defaults to 100.")
@click.option(
    "--limit", default=1000, help=" This is equivalent to a final LIMIT clause on your query. Defaults to 1000."
)
@click.pass_obj
def query(obj: Dict, query: str, source_limit: int = 100, infer_schema_limit: int = 100, limit: int = 1000) -> None:
    try:
        client = get_client(obj)
        src_lim = "all" if source_limit == -1 else source_limit  # TODO Until handled in SDK
        res = client.transformations.preview(
            query=query, source_limit=src_lim, infer_schema_limit=infer_schema_limit, limit=limit
        )
        click.echo(print_query(query, res))
    except CogniteAPIError as e:
        exit_with_cognite_api_error(e)



from typing import Dict, List, Optional

import click
from cognite.client.data_classes import TransformationJob
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.utils import (
    exit_with_cognite_api_error,
    is_id_exclusive,
    paginate,
    print_jobs,
)


def log_jobs(id_str: Optional[str], items: List[TransformationJob]) -> None:
    if id_str:
        click.echo(f"Resulting jobs for transformation with {id_str}:")
    click.echo(print_jobs(items))


@click.command(help="Show latest jobs for a given transformation")
@click.option("--id", help="List jobs by transformation id. Either this or --external-id must be specified.")
@click.option("--external-id", help="List jobs by transformation external_id. Either this or --id must be specified.")
@click.option("--limit", default=10, help="Limit for the job history, defaults to 10. Use -1 to retrieve all results.")
@click.option("-i", "--interactive", is_flag=True, help="Display only 10 jobs at a time, paging through them.")
@click.pass_obj
def jobs(obj: Dict, id: Optional[int], external_id: Optional[str], limit: int = 10, interactive: bool = False) -> None:
    client = get_client(obj)
    is_id_exclusive(id, external_id)
    try:
        id_str = None
        if id:
            id = int(id)
            id_str = f"id {id}"
        if external_id:
            id_str = f"external_id {external_id}"

        if id_str:
            click.echo(f"Listing the latest jobs for transformation with {id_str}:")
        else:
            click.echo("Listing the latest jobs for all transformations:")

        jobs = client.transformations.jobs.list(
            limit=limit, transformation_id=id, transformation_external_id=external_id
        )

        if jobs:
            if interactive:
                paginate(jobs, lambda chunk: log_jobs(id_str, chunk))
            else:
                log_jobs(id_str, jobs)
        else:
            click.echo("No jobs to list.")
    except (CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)



from typing import Dict, List

import click
from cognite.client.data_classes import Transformation
from cognite.client.exceptions import CogniteAPIError

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.utils import paginate, print_transformations


def log_transformations(transformations: List[Transformation]) -> None:
    click.echo(print_transformations(transformations))


@click.command(help="List transformations")
@click.option("--limit", default=10, help="Number of transformations to list, defaults to 10. Use -1 to list all.")
@click.option(
    "-i", "--interactive", is_flag=True, help="Display only 10 transformations at a time, then page through them."
)
@click.pass_obj
def list(obj: Dict, limit: int = 10, interactive: bool = False) -> None:
    client = get_client(obj)
    try:
        transformations = client.transformations.list(limit=limit)
        if interactive:
            paginate(transformations, log_transformations)
        else:
            log_transformations(transformations)
    except CogniteAPIError as e:
        exit(f"Cognite API error has occurred: {e}")



import sys
from typing import Dict

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteAPIKeyError


def get_client(obj: Dict) -> CogniteClient:
    api_key = obj["api_key"]
    client_id = obj["client_id"]
    client_secret = obj["client_secret"]
    token_url = obj["token_url"]
    scopes = obj["scopes"]
    audience = obj["audience"]
    cdf_project_name = obj["cdf_project_name"]
    cluster = obj["cluster"]
    base_url = f"https://{cluster}.cognitedata.com"
    if not api_key and not audience:
        scopes = scopes.strip().split(" ") if scopes else [f"https://{cluster}.cognitedata.com/.default"]
    try:
        if api_key is not None and (
            client_id is not None
            or client_secret is not None
            or token_url is not None
            or scopes is not None
            or audience is not None
        ):
            sys.exit("Please provide only API key configuration or only OAuth2 configuration.")
        elif api_key is not None:
            return CogniteClient(
                client_name="transformations_cli", api_key=api_key, base_url=base_url, project=cdf_project_name
            )
        else:
            return CogniteClient(
                base_url=base_url,
                client_name="transformations_cli",
                token_client_id=client_id,
                token_client_secret=client_secret,
                token_url=token_url,
                token_scopes=scopes,
                project=cdf_project_name,
                token_custom_args={"audience": audience} if audience else None,
            )
    except CogniteAPIKeyError as e:
        sys.exit(f"Cognite client cannot be initialised: {e}.")



import os
import sys
from typing import Dict, List, Optional, Tuple, Union

from cognite.client import CogniteClient
from cognite.client.data_classes import (
    OidcCredentials,
    Transformation,
    TransformationDestination,
    TransformationNotification,
    TransformationSchedule,
)
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError, CogniteNotFoundError

from cognite.transformations_cli.commands.deploy.transformation_types import (
    ActionType,
    AuthConfig,
    DestinationConfig,
    DestinationType,
    QueryConfig,
    ReadWriteAuthentication,
    ScheduleConfig,
    TransformationConfig,
)
from cognite.transformations_cli.commands.utils import exit_with_cognite_api_error

TupleResult = List[Tuple[str, str]]
StandardResult = List[str]


def to_transformation(conf_path: str, config: TransformationConfig, cluster: str = "europe-west1-1") -> Transformation:
    return Transformation(
        name=config.name,
        external_id=config.external_id,
        destination=to_destination(config.destination),
        conflict_mode=to_action(config.action),
        is_public=config.shared,
        ignore_null_fields=config.ignore_null_fields,
        query=to_query(conf_path, config.query),
        source_api_key=to_read_api_key(config.authentication),
        destination_api_key=to_write_api_key(config.authentication),
        source_oidc_credentials=to_read_oidc(config.authentication, cluster),
        destination_oidc_credentials=to_write_oidc(config.authentication, cluster),
    )


def to_action(action: ActionType) -> str:
    return "abort" if action == ActionType.create else action.value


def to_destination(destination: Union[DestinationType, DestinationConfig]) -> TransformationDestination:
    if isinstance(destination, DestinationConfig):
        if destination.type == DestinationType.raw:
            return TransformationDestination.raw(destination.raw_database, destination.raw_table)
        return TransformationDestination(destination.type.value)
    else:
        return TransformationDestination(destination.value)


def to_query(conf_path: str, query: Union[str, QueryConfig]) -> str:
    try:
        dir_path = os.path.dirname(conf_path)
        if isinstance(query, QueryConfig):
            sql_path = os.path.join(dir_path, query.file)
            with open(sql_path, "r") as f:
                return f.read().replace("\n", "")
        return query
    except:
        sys.exit("Please provide a valid path for sql file.")


def to_read_api_key(authentication: Union[AuthConfig, ReadWriteAuthentication]) -> Optional[str]:
    if isinstance(authentication, AuthConfig):
        return authentication.api_key
    if isinstance(authentication, ReadWriteAuthentication):
        return authentication.read.api_key
    return None


def to_write_api_key(authentication: Union[AuthConfig, ReadWriteAuthentication]) -> Optional[str]:
    if isinstance(authentication, AuthConfig):
        return authentication.api_key
    if isinstance(authentication, ReadWriteAuthentication):
        return authentication.write.api_key
    return None


def stringify_scopes(scopes: Optional[List[str]]) -> Optional[str]:
    if scopes:
        return " ".join(scopes)
    return None


def get_default_scopes(scopes: Optional[str], cluster: str) -> str:
    return scopes if scopes else f"https://{cluster}.cognitedata.com/.default"


def is_oidc_defined(auth_config: AuthConfig) -> bool:
    return (
        auth_config.client_id and auth_config.client_secret and auth_config.token_url and auth_config.cdf_project_name
    )


def get_oidc(auth_config: AuthConfig, cluster: str) -> Optional[OidcCredentials]:
    stringified_scopes = stringify_scopes(auth_config.scopes)
    scopes = stringified_scopes if auth_config.audience else get_default_scopes(stringified_scopes, cluster)
    return (
        OidcCredentials(
            client_id=auth_config.client_id,
            client_secret=auth_config.client_secret,
            scopes=scopes,
            token_uri=auth_config.token_url,
            cdf_project_name=auth_config.cdf_project_name,
            audience=auth_config.audience,
        )
        if is_oidc_defined(auth_config)
        else None
    )


def to_read_oidc(authentication: Union[AuthConfig, ReadWriteAuthentication], cluster: str) -> Optional[OidcCredentials]:
    return (
        get_oidc(authentication, cluster)
        if isinstance(authentication, AuthConfig)
        else get_oidc(authentication.read, cluster)
    )


def to_write_oidc(
    authentication: Union[AuthConfig, ReadWriteAuthentication], cluster: str
) -> Optional[OidcCredentials]:
    return (
        get_oidc(authentication, cluster)
        if isinstance(authentication, AuthConfig)
        else get_oidc(authentication.write, cluster)
    )


def to_schedule(transformation_external_id: str, schedule: Union[str, ScheduleConfig]) -> TransformationSchedule:
    if isinstance(schedule, ScheduleConfig):
        interval = schedule.interval
        is_paused = schedule.is_paused
    else:
        interval = schedule
        is_paused = False
    return TransformationSchedule(external_id=transformation_external_id, interval=interval, is_paused=is_paused)


def to_notification(transformation_external_id: str, destination: str) -> TransformationNotification:
    return TransformationNotification(transformation_external_id=transformation_external_id, destination=destination)


def get_existing_trasformation_ext_ids(client: CogniteClient, all_ext_ids: List[str]) -> List[str]:
    return [
        t.external_id
        for t in client.transformations.retrieve_multiple(external_ids=all_ext_ids, ignore_unknown_ids=True)
    ]


def get_new_transformation_ids(all_ext_ids: List[str], existig_ext_ids: List[str]) -> List[str]:
    return list(set(all_ext_ids) - set(existig_ext_ids))


def get_existing_schedules_dict(client: CogniteClient, all_ext_ids: List[str]) -> Dict[str, TransformationSchedule]:
    return {
        s.external_id: s
        for s in client.transformations.schedules.retrieve_multiple(external_ids=all_ext_ids, ignore_unknown_ids=True)
    }


def get_existing_notifications_dict(
    client: CogniteClient, all_ext_ids: List[str]
) -> Dict[str, List[TransformationNotification]]:
    existing_notifications = dict()
    for ext_id in all_ext_ids:
        notifications = client.transformations.notifications.list(transformation_external_id=ext_id, limit=-1)
        if notifications:
            existing_notifications[ext_id] = notifications
    return existing_notifications


def upsert_transformations(
    client: CogniteClient,
    transformations: List[Transformation],
    existing_ext_ids: List[str],
    new_ext_ids: List[str],
) -> Tuple[StandardResult, StandardResult, StandardResult]:
    try:
        updated = client.transformations.update([tr for tr in transformations if tr.external_id in existing_ext_ids])
        created = client.transformations.create([tr for tr in transformations if tr.external_id in new_ext_ids])
        return [], [t.external_id for t in updated], [t.external_id for t in created]
    except (CogniteDuplicatedError, CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)
    return [], [], []


def upsert_schedules(
    client: CogniteClient,
    existing_schedules_dict: Dict[str, TransformationSchedule],
    requested_schedules_dict: Dict[str, TransformationSchedule],
    existing_transformations_ext_ids: List[str],
    new_transformations_ext_ids: List[str],
) -> Tuple[StandardResult, StandardResult, StandardResult]:
    to_delete = []
    to_update = []
    to_create = []
    try:
        for ext_id in existing_transformations_ext_ids:
            if ext_id in existing_schedules_dict and ext_id not in requested_schedules_dict:
                to_delete.append(ext_id)
            elif ext_id in existing_schedules_dict:
                to_update.append(ext_id)
            elif ext_id in requested_schedules_dict:
                to_create.append(ext_id)
        to_create += [ext_id for ext_id in new_transformations_ext_ids if ext_id in requested_schedules_dict]
        client.transformations.schedules.delete(external_id=to_delete)
        client.transformations.schedules.update([requested_schedules_dict[ext_id] for ext_id in to_update])
        client.transformations.schedules.create([requested_schedules_dict[ext_id] for ext_id in to_create])
    except (CogniteDuplicatedError, CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)
    return to_delete, to_update, to_create


def upsert_notifications(
    client: CogniteClient,
    existing_notifications_dict: Dict[str, List[TransformationNotification]],
    requested_notifications_dict: Dict[str, List[TransformationNotification]],
    existing_transformations_ext_ids: List[str],
    new_transformations_ext_ids: List[str],
) -> Tuple[TupleResult, TupleResult, TupleResult]:
    try:
        to_delete = dict()
        to_create = []
        for ext_id in new_transformations_ext_ids:
            if ext_id in requested_notifications_dict:
                to_create += requested_notifications_dict[ext_id]

        for ext_id in existing_transformations_ext_ids:
            existing_notif = existing_notifications_dict.get(ext_id, [])
            requested_notif = requested_notifications_dict.get(ext_id, [])
            existing_destinations = [e.destination for e in existing_notif]
            requested_destinations = [e.destination for e in requested_notif]

            to_delete.update(
                {e.id: (ext_id, e.destination) for e in existing_notif if e.destination not in requested_destinations}
            )
            to_create += [e for e in requested_notif if e.destination not in existing_destinations]
        client.transformations.notifications.delete(list(to_delete.keys()))
        client.transformations.notifications.create(to_create)
        return (
            [to_delete[key] for key in to_delete],
            [],
            [(n.transformation_external_id, n.destination) for n in to_create],
        )
    except (CogniteDuplicatedError, CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)
    return [], [], []



import sys
from typing import Dict, Optional

import click
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.utils import (
    exit_with_cognite_api_error,
    get_transformation,
    is_id_exclusive,
    print_jobs,
    print_metrics,
    print_notifications,
    print_sql,
    print_transformations,
)


@click.command(help="Show detalis of a transformation")
@click.option("--id", help="The id of the transformation to show. Either this or --external-id can be specified.")
@click.option(
    "--external-id", help="The external_id of the transformation to show. Either this or --id can be specified."
)
@click.option("--job-id", help="The id of the job to show. Include this to show job details.")
@click.pass_obj
def show(obj: Dict, id: Optional[int], external_id: Optional[str], job_id: Optional[int]) -> None:
    client = get_client(obj)
    is_id_exclusive(id, external_id)
    if not (id or external_id or job_id):
        click.echo("Please provide id, external_id or job_id")
        sys.exit(1)
    try:
        tr = None
        job = None
        if id or external_id:
            # TODO Investigate why id requires type casting as it doesn't in "jobs command"
            id = int(id) if id else None
            tr = get_transformation(client=client, id=id, external_id=external_id)
            click.echo("Transformation details:")
            click.echo(print_transformations([tr]))
            notifications = client.transformations.notifications.list(
                transformation_id=id, transformation_external_id=external_id, limit=-1
            )
            if tr.query:
                click.echo("SQL Query:")
                click.echo(print_sql(tr.query))
            if notifications:
                click.echo("Notifications:")
                click.echo(print_notifications(notifications))
        if job_id:
            click.echo()
            job_id = int(job_id)
            job = client.transformations.jobs.retrieve(id=int(job_id))
            metrics = [
                m
                for m in client.transformations.jobs.list_metrics(id=job_id)
                if m.name != "requestsWithoutRetries" and m.name != "requests"
            ]
            click.echo("Job details:")
            click.echo(print_jobs([job]))
            click.echo("SQL Query:")
            click.echo(print_sql(job.query))
            if job.status == "Failed":
                click.echo(f"Error Details: {job.error}")
            if metrics:
                click.echo("Progress:")
                click.echo(print_metrics(metrics))
    except (CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)



import sys
import time
from typing import Dict, Optional

import click
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.utils import (
    exit_with_cognite_api_error,
    is_id_exclusive,
    is_id_provided,
    print_jobs,
    print_metrics,
    print_sql,
)


@click.command(help="Start and/or watch transformation jobs")
@click.option("--id", help="The id of the transformation to run. Either this or --external-id must be specified.")
@click.option(
    "--external-id", help="The external_id of the transformation to run. Either this or --id must be specified."
)
@click.option("--watch", is_flag=True, help="Wait until job has completed.")
@click.option(
    "--watch-only",
    is_flag=True,
    help="Do not start a transformation job, only watch the most recent job for completion.",
)
@click.option(
    "--time-out",
    default=(12 * 60 * 60),
    help="Maximum amount of time to wait for job to complete in seconds, 12 hours by default.",
)
@click.pass_obj
def run(
    obj: Dict,
    id: Optional[int],
    external_id: Optional[str],
    watch: bool = False,
    watch_only: bool = False,
    time_out: int = (12 * 60 * 60),
) -> None:
    client = get_client(obj)
    is_id_provided(id, external_id)
    is_id_exclusive(id, external_id)
    try:
        job = None
        # TODO Investigate why id requires type casting as it doesn't in "jobs command"
        id = int(id) if id else None
        if not watch_only:
            duration_start = time.time()
            job = client.transformations.run(
                transformation_id=id, transformation_external_id=external_id, wait=watch, timeout=time_out
            )
            duration_end = time.time()
        else:
            jobs = client.transformations.jobs.list(transformation_id=id, transformation_external_id=external_id)

            duration_start = time.time()
            job = jobs[0].wait(timeout=time_out) if jobs else None
            duration_end = time.time()
        if job:
            metrics = [
                m
                for m in client.transformations.jobs.list_metrics(id=job.id)
                if m.name != "requestsWithoutRetries" and m.name != "requests"
            ]
            click.echo("Job details:")
            click.echo(print_jobs([job]))
            click.echo("SQL Query:")
            click.echo(print_sql(job.query))
            if job.status == "Failed":
                click.echo(f"Job Failed, error details: {job.error}")
            if metrics:
                click.echo("Progress:")
                click.echo(print_metrics(metrics))
            if watch_only or watch:
                if job.status == "Failed":
                    # Error already been printed so just exit.
                    sys.exit(1)
                if duration_end - duration_start > (time_out + 1) and job.status != "Completed":
                    click.echo(f"Transformation job runtime exceeds the provided timeout: {time_out} seconds")
                    sys.exit(1)
    # Handle AttributeError because SDK fails here:
    # transformation_id = self.retrieve(external_id=transformation_external_id).id
    # with "AttributeError: 'NoneType' object has no attribute 'id'"
    # when called with invalid external_id
    # We mask that here with this trick until solved.
    except AttributeError:
        click.echo("Cognite API error has occurred: Transformation not found.")
        sys.exit(1)
    except (CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)



import glob
import os
from typing import Dict, List, Optional, Union

from cognite.extractorutils.configtools import load_yaml
from regex import regex

from cognite.transformations_cli.commands.deploy.transformation_types import (
    AuthConfig,
    DestinationConfig,
    DestinationType,
    ReadWriteAuthentication,
    TransformationConfig,
    TransformationConfigError,
)
from cognite.transformations_cli.commands.deploy.transformation_types_legacy import TransformationConfigLegacy


def _validate_destination_type(external_id: str, destination_type: DestinationConfig) -> None:
    if destination_type.type == DestinationType.raw and (
        destination_type.raw_database is None or destination_type.raw_table is None
    ):
        raise Exception(f"Raw destination type requires database and table properties to be set: {external_id}")
    return None


def _validate_exclusive_auth(external_id: str, auth: Optional[AuthConfig]) -> None:
    if (
        auth
        and auth.api_key
        and (auth.client_id or auth.client_secret or auth.cdf_project_name or auth.scopes or auth.token_url)
    ):
        raise Exception(f"Please provide only one of api-key or OAuth2 credentials: {external_id}")
    return None


def _validate_auth(external_id: str, auth_config: Union[AuthConfig, ReadWriteAuthentication]) -> None:
    if isinstance(auth_config, AuthConfig):
        _validate_exclusive_auth(external_id, auth_config)
    if isinstance(auth_config, ReadWriteAuthentication):
        _validate_exclusive_auth(external_id, auth_config.read)
        _validate_exclusive_auth(external_id, auth_config.write)


def _validate_config(config: TransformationConfig) -> None:
    _validate_destination_type(config.external_id, config.destination)
    _validate_auth(config.external_id, config.authentication)


def _parse_transformation_config(path: str) -> TransformationConfig:
    r = regex.compile(r"^legacy:\s*true\s*$", flags=regex.MULTILINE | regex.IGNORECASE)
    with open(path) as f:
        data = f.read()
        if r.search(data) is not None:
            legacy = load_yaml(data, TransformationConfigLegacy, case_style="camel")
            return legacy.to_new()
        else:
            return load_yaml(data, TransformationConfig, case_style="camel")


def parse_transformation_configs(base_dir: Optional[str]) -> Dict[str, TransformationConfig]:
    if base_dir is None:
        base_dir = "."

    if os.path.isdir(base_dir) is False:
        raise TransformationConfigError(f"Transformation root folder not found: {base_dir}")

    transformations: Dict[str, TransformationConfig] = dict()
    yaml_paths: List[str] = glob.glob(f"{base_dir}/**/*.yaml", recursive=True) + glob.glob(
        f"{base_dir}/**/*.yml", recursive=True
    )

    for file_path in yaml_paths:
        try:
            parsed_conf = _parse_transformation_config(file_path)
            # This will raise exceptions if invalid
            _validate_config(parsed_conf)
            transformations[file_path] = parsed_conf
        except Exception as e:
            raise TransformationConfigError(
                f"Failed to parse transformation config, please check that you conform required fields and format: {e}"
            )
    return transformations



from dataclasses import dataclass, field
from os import environ
from typing import List, Optional, Union

from cognite.transformations_cli.commands.deploy.transformation_types import (
    ActionType,
    AuthConfig,
    DestinationConfig,
    DestinationType,
    QueryConfig,
    ReadWriteAuthentication,
    TransformationConfig,
    TransformationConfigError,
)


# YamlDotNet is case insensitive on enums, for full backwards compatibility that
# needs to be reflected here
def legacy_destination_type_to_new(input: str) -> DestinationType:
    input_low = input.lower()
    for en in DestinationType:
        clean_name = en.name.replace("_", "")
        if clean_name == input_low:
            return en
    raise TransformationConfigError(f"Invalid destination type: {input}")


def legacy_action_to_new(input: str) -> ActionType:
    input_low = input.lower()
    for en in ActionType:
        if en.name == input_low:
            return en
    raise TransformationConfigError(f"Invalid action type: {input}")


@dataclass
class ReadWriteApiKeyLegacy:
    read: str
    write: str


@dataclass
class AuthConfigLegacy:
    client_id: Optional[str]
    client_secret: Optional[str]
    token_url: Optional[str]
    scopes: Optional[List[str]]
    cdf_project_name: Optional[str]
    audience: Optional[str]

    def to_new(self) -> AuthConfig:
        return AuthConfig(
            api_key=None,
            client_id=environ[self.client_id] if self.client_id is not None else None,
            audience=self.audience,
            client_secret=environ[self.client_secret] if self.client_secret is not None else None,
            token_url=self.token_url,
            scopes=self.scopes,
            cdf_project_name=self.cdf_project_name,
        )


@dataclass
class ReadWriteAuthConfigLegacy:
    read: AuthConfigLegacy
    write: AuthConfigLegacy


@dataclass
class DestinationLegacy:
    type: str
    raw_database: Optional[str] = None
    raw_table: Optional[str] = None

    def to_new(self) -> DestinationConfig:
        new_type = legacy_destination_type_to_new(self.type)
        return DestinationConfig(new_type, self.raw_database, self.raw_table)


@dataclass
class TransformationConfigLegacy:
    external_id: str
    name: str
    query: str
    authentication: Union[AuthConfigLegacy, ReadWriteAuthConfigLegacy, None]
    api_key: Union[str, ReadWriteApiKeyLegacy, None]
    schedule: Optional[str]
    destination: DestinationLegacy
    notifications: List[str] = field(default_factory=list)
    shared: bool = False
    ignore_null_fields: bool = True
    action: str = "upsert"
    legacy: bool = True

    def to_new(self) -> TransformationConfig:
        query = QueryConfig(file=self.query)
        auth = ReadWriteAuthentication(
            read=AuthConfig(None, None, None, None, None, None, None),
            write=AuthConfig(None, None, None, None, None, None, None),
        )

        if isinstance(self.authentication, ReadWriteAuthConfigLegacy):
            if self.authentication.read is not None:
                auth.read = self.authentication.read.to_new()
            if self.authentication.write is not None:
                auth.write = self.authentication.write.to_new()
        elif isinstance(self.authentication, AuthConfigLegacy):
            auth.read = auth.write = self.authentication.to_new()

        if isinstance(self.api_key, ReadWriteApiKeyLegacy):
            if self.api_key.read is not None:
                auth.read.api_key = environ[self.api_key.read]
            if self.api_key.write is not None:
                auth.write.api_key = environ[self.api_key.write]
        elif isinstance(self.api_key, str):
            auth.read.api_key = auth.write.api_key = environ[self.api_key]

        destination = self.destination.to_new()
        action = legacy_action_to_new(self.action)

        return TransformationConfig(
            external_id=self.external_id,
            name=self.name,
            query=query,
            authentication=auth,
            destination=destination,
            notifications=self.notifications,
            shared=self.shared,
            ignore_null_fields=self.ignore_null_fields,
            action=action,
            schedule=self.schedule,
            legacy=True,
        )



from typing import Dict, Optional

import click
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.utils import exit_with_cognite_api_error, is_id_exclusive, is_id_provided


@click.command(help="Delete a transformation")
@click.option("--id", help="The id of the transformation to show. Either this or --external-id must be specified.")
@click.option(
    "--external-id", help="The external_id of the transformation to show. Either this or --id must be specified."
)
@click.option("--delete-schedule", is_flag=False, help="Delete the schedule before deleting the transformation.")
@click.pass_obj
def delete(obj: Dict, id: Optional[int], external_id: Optional[str], delete_schedule: bool = False) -> None:
    client = get_client(obj)
    id = int(id) if id else None
    is_id_provided(id, external_id)
    is_id_exclusive(id, external_id)
    try:
        if delete_schedule:
            client.transformations.schedules.delete(external_id=external_id, id=id, ignore_unknown_ids=True)
        client.transformations.delete(external_id=external_id, id=id)
        if id:
            click.echo(f"Successfully deleted the transformation with id {id}.")
        else:
            click.echo(f"Successfully deleted the transformation with external id {external_id}.")
    except (CogniteNotFoundError, CogniteAPIError) as e:
        exit_with_cognite_api_error(e)



from typing import Optional

import click
from click import Context

from cognite.transformations_cli import __version__
from cognite.transformations_cli.commands.delete import delete
from cognite.transformations_cli.commands.deploy.deploy import deploy
from cognite.transformations_cli.commands.jobs import jobs
from cognite.transformations_cli.commands.list import list
from cognite.transformations_cli.commands.query import query
from cognite.transformations_cli.commands.run import run
from cognite.transformations_cli.commands.show import show


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(prog_name="transformations_cli", version=__version__)
@click.option(
    "--cluster",
    default="europe-west1-1",
    help="The CDF cluster where Transformations is hosted (e.g. greenfield, europe-west1-1)",
    envvar="TRANSFORMATIONS_CLUSTER",
)
@click.option(
    "--api-key",
    help="API key to interact with transformations API. Provide this or make sure to set 'TRANSFORMATIONS_API_KEY' environment variable if you want to authenticate with API keys.",
    envvar="TRANSFORMATIONS_API_KEY",
)
@click.option(
    "--client-id",
    help="Client ID to interact with transformations API. Provide this or make sure to set 'TRANSFORMATIONS_CLIENT_ID' environment variable if you want to authenticate with OAuth2.",
    envvar="TRANSFORMATIONS_CLIENT_ID",
)
@click.option(
    "--client-secret",
    help="Client secret to interact with transformations API. Provide this or make sure to set 'TRANSFORMATIONS_CLIENT_SECRET' environment variable if you want to authenticate with OAuth2.",
    envvar="TRANSFORMATIONS_CLIENT_SECRET",
)
@click.option(
    "--token-url",
    help="Token URL to interact with transformations API. Provide this or make sure to set 'TRANSFORMATIONS_TOKEN_URL' environment variable if you want to authenticate with OAuth2.",
    envvar="TRANSFORMATIONS_TOKEN_URL",
)
@click.option(
    "--scopes",
    help="Scopes to interact with transformations API, relevant for OAuth2 authentication method. 'TRANSFORMATIONS_SCOPES' environment variable can be used instead.",
    envvar="TRANSFORMATIONS_SCOPES",
)
@click.option(
    "--audience",
    help="Audience to interact with transformations API, relevant for OAuth2 authentication method. 'TRANSFORMATIONS_AUDIENCE' environment variable can be used instead.",
    envvar="TRANSFORMATIONS_AUDIENCE",
)
@click.option(
    "--cdf-project-name",
    help="Project to interact with transformations API, 'TRANSFORMATIONS_PROJECT' environment variable can be used instead. Required for OAuth2 and optional for api-keys.",
    envvar="TRANSFORMATIONS_PROJECT",
)
@click.pass_context
def transformations_cli(
    context: Context,
    cluster: str = "europe-west1-1",
    api_key: Optional[str] = None,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    token_url: Optional[str] = None,
    scopes: Optional[str] = None,
    audience: Optional[str] = None,
    cdf_project_name: Optional[str] = None,
) -> None:
    context.obj = {
        "cluster": cluster,
        "api_key": api_key,
        "client_id": client_id,
        "client_secret": client_secret,
        "token_url": token_url,
        "scopes": scopes,
        "audience": audience,
        "cdf_project_name": cdf_project_name,
    }


transformations_cli.add_command(deploy)
transformations_cli.add_command(run)
transformations_cli.add_command(query)
transformations_cli.add_command(list)
transformations_cli.add_command(show)
transformations_cli.add_command(jobs)
transformations_cli.add_command(delete)



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union


class DestinationType(Enum):
    assets = "assets"
    timeseries = "timeseries"
    asset_hierarchy = "asset_hierarchy"
    events = "events"
    datapoints = "datapoints"
    string_datapoints = "string_datapoints"
    sequences = "sequences"
    files = "files"
    labels = "labels"
    relationships = "relationships"
    raw = "raw"
    data_sets = "data_sets"


class ActionType(Enum):
    create = "create"
    abort = "abort"
    update = "update"
    upsert = "upsert"
    delete = "delete"


@dataclass
class AuthConfig:
    api_key: Optional[str]
    client_id: Optional[str]
    client_secret: Optional[str]
    token_url: Optional[str]
    scopes: Optional[List[str]]
    cdf_project_name: Optional[str]
    audience: Optional[str]


@dataclass
class ReadWriteAuthentication:
    read: AuthConfig
    write: AuthConfig


@dataclass
class DestinationConfig:
    """
    Valid type values are: assets, asset_hierarchy, events, timeseries, datapoints, string_datapoints, raw (needs database and table)
    """

    type: DestinationType
    raw_database: Optional[str] = None
    raw_table: Optional[str] = None


@dataclass
class QueryConfig:
    file: str


@dataclass
class ScheduleConfig:
    interval: str
    is_paused: bool = False


@dataclass
class TransformationConfig:
    """
    Master configuration class of a transformation
    """

    external_id: str
    name: str
    query: Union[str, QueryConfig]
    authentication: Union[AuthConfig, ReadWriteAuthentication]
    schedule: Optional[Union[str, ScheduleConfig]]
    destination: Union[DestinationType, DestinationConfig]
    notifications: List[str] = field(default_factory=list)
    shared: bool = True
    ignore_null_fields: bool = True
    action: ActionType = ActionType.upsert
    legacy: bool = False


class TransformationConfigError(Exception):
    """Exception raised for config parser

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['cognite',
 'cognite.transformations_cli',
 'cognite.transformations_cli.commands',
 'cognite.transformations_cli.commands.deploy']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'cognite-extractor-utils>=1.5.2,<2.0.0',
 'cognite-sdk>=2.38.1,<3.0.0',
 'regex>=2021.11.10,<2022.0.0',
 'sqlparse>=0.4.2,<0.5.0',
 'tabulate>=0.8.9,<0.9.0',
 'types-retry>=0.1.5,<0.2.0',
 'types-tabulate>=0.8.3,<0.9.0']

entry_points = \
{'console_scripts': ['transformations-cli = '
                     'cognite.transformations_cli.__main__:main']}

setup_kwargs = {
    'name': 'cognite-transformations-cli',
    'version': '1.2.6',
    'description': 'A CLI for the Transformations service in CDF',
    'long_description': '<a href="https://cognite.com/">\n    <img src="https://github.com/cognitedata/cognite-python-docs/blob/master/img/cognite_logo.png" alt="Cognite logo" title="Cognite" align="right" height="80" />\n</a>\n\nCognite Python `transformations-cli`\n================================\n[![Build Status](https://github.com/cognitedata/transformations-cli/workflows/release/badge.svg)](https://github.com/cognitedata/transformations-cli/actions)\n[![Documentation Status](https://readthedocs.com/projects/cognite-transformations-cli/badge/?version=latest)](https://cognite-transformations-cli.readthedocs-hosted.com/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/cognitedata/transformations-cli/branch/main/graph/badge.svg?token=PSkli74vvX)](https://codecov.io/gh/cognitedata/transformations-cli)\n[![PyPI version](https://badge.fury.io/py/cognite-transformations-cli.svg)](https://pypi.org/project/cognite-transformations-cli)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/cognite-transformations-cli)\n[![License](https://img.shields.io/github/license/cognitedata/python-extractor-utils)](LICENSE)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n# Transformations CLI\n\nThe Transormations CLI is a replacement for [jetfire-cli](https://github.com/cognitedata/jetfire-cli) rewritten on top\nof the new Python SDK for Transformations.\n\n### CLI Documentation\n\nDocumentation for CLI is hosted [here](https://cognite-transformations-cli.readthedocs-hosted.com/en/latest/).\n\n### GitHub Action\n\n`transformations-cli` also provides a GitHub Action which can be used to deploy transformations. You can find the documentation for transformations-cli GitHub Action [here](githubaction.md).\n\n\n### GitHub Action Migration: jetfire-cli@v2 to transformations-cli@main\n\nIf you\'ve already used the old `jetfire-cli` in a GitHub Action we recommend you migrate to the new GitHub Action. You can find the migration guide [here](migrationguide.md).\n\n### Contributing\n\nWe use [poetry](https://python-poetry.org) to manage dependencies and to administrate virtual environments. To develop\n`transformations-cli`, follow the following steps to set up your local environment:\n\n 1. Install poetry: (add `--user` if desirable)\n    ```\n    $ pip install poetry\n    ```\n 2. Clone repository:\n    ```\n    $ git clone git@github.com:cognitedata/transformations-cli.git\n    ```\n 3. Move into the newly created local repository:\n    ```\n    $ cd transformations-cli\n    ```\n 4. Create virtual environment and install dependencies:\n    ```\n    $ poetry install\n    ```\n\nAll code must pass [black](https://github.com/ambv/black) and [isort](https://github.com/timothycrosley/isort) style\nchecks to be merged. It is recommended to install pre-commit hooks to ensure this locally before commiting code:\n\n```\n$ poetry run pre-commit install\n```\n\nTo publish a new version change the version in `cognite/transformations_cli/__init__.py` and `pyproject.toml`. Also, remember to add an entry to `CHANGELOG`.\n\nThis project adheres to the [Contributor Covenant v2.0](https://www.contributor-covenant.org/version/2/0/code_of_conduct/)\nas a code of conduct.\n\n\n',
    'author': 'Mathias Lohne',
    'author_email': 'mathias.lohne@cognite.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cognitedata/transformations-cli',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, Union

import click
from cognite.client import CogniteClient
from cognite.client.data_classes import OidcCredentials, Transformation

from cognite.transformations_cli.clients import get_client
from cognite.transformations_cli.commands.deploy.transformation_config import (
    TransformationConfigError,
    parse_transformation_configs,
)
from cognite.transformations_cli.commands.deploy.transformations_api import (
    StandardResult,
    TupleResult,
    get_existing_notifications_dict,
    get_existing_schedules_dict,
    get_existing_trasformation_ext_ids,
    get_new_transformation_ids,
    to_notification,
    to_schedule,
    to_transformation,
    upsert_notifications,
    upsert_schedules,
    upsert_transformations,
)


def verify_oidc_credentials(type: str, credentials: OidcCredentials, cluster: str) -> None:
    token_inspect = None
    base_url = f"https://{cluster}.cognitedata.com"
    try:
        client = CogniteClient(
            base_url=base_url,
            client_name="transformations-cli-credentials-test",
            project=credentials.cdf_project_name,
            token_url=credentials.token_uri,
            token_client_id=credentials.client_id,
            token_client_secret=credentials.client_secret,
            token_scopes=credentials.scopes,
            token_custom_args={"audience": credentials.audience} if credentials.audience else None,
        )
        token_inspect = client.iam.token.inspect()
    except Exception as ex:
        raise TransformationConfigError(f"Credentials for {type} failed to validate: {str(ex)}")

    if not next((x for x in token_inspect.projects if x.url_name == credentials.cdf_project_name), None):
        raise TransformationConfigError(
            f"Credentials for {type} failed to validate: they lack projects:list and groups:list in project {credentials.cdf_project_name}"
        )


def verify_credentials(t: Transformation, cluster: str) -> None:
    if t.has_destination_oidc_credentials:
        verify_oidc_credentials(f"{t.name} write", t.destination_oidc_credentials, cluster)
    if t.has_source_oidc_credentials:
        verify_oidc_credentials(f"{t.name} read", t.source_oidc_credentials, cluster)


def print_results(
    resource_type: str, action: str, results: Union[StandardResult, TupleResult], debug: bool = False
) -> None:
    if results:
        click.echo(f"Number of {resource_type}s {action}d: {len(results)}")
        if debug:
            click.echo(f"List of {resource_type}s {action}d:")
            click.echo(results)
    return None


@click.command(help="Deploy a set of transformations from a directory")
@click.argument(
    "path",
    default=".",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Print updated, created, deleted external IDs ",
)
@click.pass_obj
def deploy(obj: Dict, path: str, debug: bool = False) -> None:
    """
        Deploy a set of transformations from a directory
    Args:
        path (str): Root directory for transformations
    """
    click.echo("Deploying transformation...")
    try:
        client = get_client(obj)
        cluster = obj["cluster"]
        transformation_configs = parse_transformation_configs(path)
        transformations = [
            to_transformation(conf_path, transformation_configs[conf_path], cluster)
            for conf_path in transformation_configs
        ]
        transformations_ext_ids = [t.external_id for t in transformation_configs.values()]

        for t in transformations:
            verify_credentials(t, cluster)

        existing_transformations_ext_ids = get_existing_trasformation_ext_ids(client, transformations_ext_ids)
        new_transformation_ext_ids = get_new_transformation_ids(
            transformations_ext_ids, existing_transformations_ext_ids
        )

        _, updated_transformations, created_transformations = upsert_transformations(
            client, transformations, existing_transformations_ext_ids, new_transformation_ext_ids
        )

        print_results("transformation", "update", updated_transformations, debug)
        print_results("transformation", "create", created_transformations, debug)

        existing_schedules_dict = get_existing_schedules_dict(client, transformations_ext_ids)
        existing_notifications_dict = get_existing_notifications_dict(client, transformations_ext_ids)

        requested_schedules_dict = {
            t.external_id: to_schedule(t.external_id, t.schedule) for t in transformation_configs.values() if t.schedule
        }

        deleted_schedules, updated_schedules, created_schedules = upsert_schedules(
            client,
            existing_schedules_dict,
            requested_schedules_dict,
            existing_transformations_ext_ids,
            new_transformation_ext_ids,
        )

        print_results("schedule", "delete", deleted_schedules, debug)
        print_results("schedule", "update", updated_schedules, debug)
        print_results("schedule", "create", created_schedules, debug)

        requested_notifications_dict = dict()
        for t in transformation_configs.values():
            if t.notifications:
                notifs = [to_notification(t.external_id, dest) for dest in t.notifications]
                requested_notifications_dict[t.external_id] = notifs

        deleted_notifications, _, created_notifications = upsert_notifications(
            client,
            existing_notifications_dict,
            requested_notifications_dict,
            existing_transformations_ext_ids,
            new_transformation_ext_ids,
        )

        print_results("notification", "delete", deleted_notifications, debug)
        print_results("notification", "create", created_notifications, debug)
    except TransformationConfigError as e:
        exit(e.message)



import datetime
import sys
from typing import Callable, List, Optional, TypeVar

import click
import sqlparse
from cognite.client import CogniteClient
from cognite.client.data_classes import (
    RawTable,
    Transformation,
    TransformationDestination,
    TransformationJob,
    TransformationJobMetric,
    TransformationNotification,
    TransformationPreviewResult,
)
from tabulate import tabulate


def get_transformation(client: CogniteClient, id: Optional[int], external_id: Optional[str]) -> Transformation:
    tr = client.transformations.retrieve(external_id=external_id, id=id)
    if tr:
        return tr
    msg = "external_id" if external_id else "id"
    sys.exit(
        f"Cognite API error has occurred: Transformation with {msg} {external_id if external_id else id} not found."
    )


def is_id_exclusive(id: Optional[int], externalId: Optional[str], should_exit: bool = True) -> bool:
    if id and externalId:
        if should_exit:
            sys.exit("Please only provide one of id and external id.")
        return False
    return True


def is_id_provided(id: Optional[int], externalId: Optional[str], should_exit: bool = True) -> bool:
    if not id and not externalId:
        if should_exit:
            sys.exit("Please provide one of id and external id.")
        return False
    return True


def exit_with_cognite_api_error(e: Exception) -> None:
    sys.exit(f"Cognite API error has occurred: {e}")
    return None  # To suppress mypy error


def get_table(destination: TransformationDestination) -> str:
    return destination.table if isinstance(destination, RawTable) else ""


def get_database(destination: TransformationDestination) -> str:
    return destination.database if isinstance(destination, RawTable) else ""


def print_transformations(transformation: List[Transformation]) -> str:
    # TODO print last_finished_job and running_job, schedule, blocked details when implemented in SDK
    list_columns = [["Name", "ID", "External ID", "Destination", "Database", "Table", "Action"]]
    return tabulate(
        list_columns
        + [
            [
                t.name,
                t.id,
                t.external_id,
                t.destination.type,
                get_database(t.destination),
                get_table(t.destination),
                t.conflict_mode,
            ]
            for t in transformation
        ],
        headers="firstrow",
        tablefmt="rst",
    )


def print_query(query: str, result: TransformationPreviewResult) -> str:
    print_res = f"Query:\n{print_sql(query)}\n"
    schema = result.schema
    results = result.results
    if schema:
        print_res += "\nSchema:\n"
        schema_content = [["name", "type", "nullable"]] + [[s.name, s.type.type, s.nullable] for s in schema]
        print_res += tabulate(schema_content, headers="firstrow", tablefmt="rst") + "\n"
    if results:
        print_res += "\nResults:\n"
        res_content = [[key for key in results[0]]]
        for result in results:
            res_content.append([result[key] for key in res_content[0]])
        print_res += tabulate(res_content, headers="firstrow", tablefmt="rst")
    return print_res


def print_sql(query: str) -> str:
    return sqlparse.format(query, reindent=True, keyword_case="upper")


def format_datetime(ms: Optional[int]) -> str:
    return str(datetime.datetime.fromtimestamp(ms / 1e3).replace(microsecond=0)) if ms else ""


def print_notifications(notifications: List[TransformationNotification]) -> bool:
    list_columns = [["ID", "Transformation ID", "Transformation External ID", "Destination", "Created Time"]]
    return tabulate(
        list_columns
        + [
            [n.id, n.transformation_id, n.transformation_external_id, n.destination, format_datetime(n.created_time)]
            for n in notifications
        ],
        headers="firstrow",
        tablefmt="rst",
    )


def print_metrics(metrics: List[TransformationJobMetric]) -> str:
    list_columns = [["Timestamp", "Metric Name", "Count"]]
    return tabulate(
        list_columns + [[format_datetime(m.timestamp), m.name, m.count] for m in metrics],
        headers="firstrow",
        tablefmt="rst",
    )


def get_job_duration(ms_start: Optional[int], ms_finished: Optional[int]) -> str:
    # TODO due to a bug in the backend, sometimes the finished_time is smaller than started_time. We use abs until fixed.
    return str(datetime.timedelta(milliseconds=abs(ms_finished - ms_start))) if ms_start and ms_finished else ""


def print_jobs(jobs: List[TransformationJob]) -> str:
    list_columns = [["ID", "Transformation ID", "Created Time", "Started Time", "Finished Time", "Duration", "Status"]]
    return tabulate(
        list_columns
        + [
            [
                j.id,
                j.transformation_id,
                format_datetime(j.created_time),
                format_datetime(j.started_time),
                format_datetime(j.finished_time),
                get_job_duration(j.started_time, j.finished_time),
                j.status,
            ]
            for j in jobs
        ],
        headers="firstrow",
        tablefmt="rst",
    )


T = TypeVar("T")


def paginate(items: List[T], action: Callable[[List[T]], None]) -> None:
    chunks = [items[i : i + 10] for i in range(0, len(items), 10)]
    for chunk in chunks:
        click.clear()
        action(chunk)
        input("Press Enter to continue")
